#
# コールバック関数
#
from concurrent.futures import Future

import usb
import usbtmc


def did_send_command(command_future: Future[None]):
    """コマンド送信完了時のコールバック

    Args:
        future (Future): エグゼキュータに投げたFutureオブジェクト
    """
    try:
        _ = command_future.result()
    except (usbtmc.usbtmc.UsbtmcException, usb.core.USBError) as unexpected_error:
        print(f"Unexpected USB error occured during sending command: {unexpected_error}")


def did_receive_response(response_future: Future[bytes]):
    """受信完了時のコールバック

    Args:
        future (Future): エグゼキュータに投げたFutureオブジェクト
    """

    # とりあえず受け取る
    try:
        result = response_future.result()
    except usb.core.USBTimeoutError:
        print("Response timed out. please check command syntax or connection.")
        return
    except (usbtmc.usbtmc.UsbtmcException, usb.core.USBError) as unexpected_error:
        print(f"Unexpected USB error during waiting response: {unexpected_error}")
        return

    # 文字列への変換を試みる ダメなら16進数でダンプする
    # TODO: フォーマットを変更したり既知のコマンドについては決まった対応にしたり
    result_repr = "<< Unknown format >>"
    try:
        result_repr = result.decode()
    except UnicodeError:
        # 16byteごと分割
        separated_bytes = [result[i * 16: (i + 1) * 16] for i in range(int(len(result) / 16))]
        result_repr = "\n".join([f"{index:04X}: " + " ".join([f"{n:02X}" for n in line]) for index, line in enumerate(separated_bytes)])
    finally:
        print(result_repr, end="", flush=True)



#
# USBデバイス選択
#


from typing import Optional

import usb
import usbtmc


def choose_tmc_device() -> Optional[usb.core.Device]:
    """通信対象のTMCデバイスをユーザに選択させる

    Returns:
        Optional[usb.core.Device]: 選択されたデバイス

    Note:
        メインスレッドでの実行を想定しています。
        有効なデバイスが選択されるか、SIGINT(^C), EOF(^D) が入力されるまでループします。
        デバイスが選択されないままループを抜けた場合はNoneが返ります。
    """

    selected_device: Optional[usb.core.Device] = None
    end_request: bool = False
    while not end_request:
        # 一覧表示
        print("Connected USB-TMC devices:")
        connected_tmc_devices = usbtmc.list_devices()
        for index, tmc_device in enumerate(connected_tmc_devices):
            # デバイスの基本的な情報を表示
            device_info_str = create_device_info_string(tmc_device)
            print(f"  [{index}] {device_info_str}")

        # 選択させる
        print("Enter the index of the device you want to communicate with.")
        print("Type 'r' to reload, ^C or ^D to abort.")
        try:
            command = ""
            while command == "":
                command = input("> ")
        except (KeyboardInterrupt, EOFError):
            print("abort")
            end_request = True
            continue

        if command == "r":
            print("Reloading..")
            continue

        try:
            target_device_index = int(command)
            if target_device_index < 0 or target_device_index >= len(connected_tmc_devices):
                print("Invalid index")
                continue

            selected_device = connected_tmc_devices[target_device_index]
            end_request = True
        except ValueError:
            pass

    return selected_device


def create_device_info_string(device: usb.core.Device) -> str:
    """USBデバイスの基本的な情報を取得し、文字列形式で返す

    Args:
        device (usb.Device): 対象のデバイス

    Returns:
        str: デバイス情報を表す文字列

    Note:
        idVendor, idProductを調べ、16進数で表現します。
        またiProductからディスクリプタを取得し、製品情報の取得を試みます。
    """
    # デバイスの基本的な情報を取得
    device_idvendor: int = device.idVendor      # type:ignore
    device_idproduct: int = device.idProduct    # type:ignore
    device_iproduct: int = device.iProduct      # type:ignore
    try:
        product_info_str = usb.util.get_string(device, device_iproduct) or "<< No device info >>"
    except usb.USBError:
        product_info_str = "<< Failed to fetch device info >>"

    # 表示
    return f"{product_info_str} ({device_idvendor:04X}:{device_idproduct:04X})"



#
# tmc_shell CLI
#
import sys
from concurrent.futures import Future, ThreadPoolExecutor
from typing import Optional

import usb
import usbtmc

from . import version
from .device_chooser import choose_tmc_device
from .tmc_callbacks import did_receive_response, did_send_command


def main() -> int:
    print(f"USB-TMC shell {version}")

    # 通信対象の選択
    device: Optional[usb.core.Device] = choose_tmc_device()
    if device is None:
        return 1

    # 通信路の確立
    print("Establish connection...")
    instrument = usbtmc.Instrument(device)
    instrument.timeout = 5
    try:
        instrument.open()
    except (usbtmc.usbtmc.UsbtmcException, usb.core.USBError) as e:
        print(f"Failed to establish connection with TMC device: {e}")
        return 1

    # TMCデバイスとの通信を担うエグゼキュータ
    tmc_command_executor = ThreadPoolExecutor()

    # コマンドのやりとり
    end_request: bool = False
    while not end_request:
        # コマンド入力 ^C, ^Dで終了
        try:
            command: str = ""
            while command == "":
                command = input(">>> ")
        except (KeyboardInterrupt, EOFError):
            print("abort")
            end_request = True
            continue

        # コマンドをエグゼキュータに投げる
        command_send_future: Future = tmc_command_executor.submit(instrument.write, command)
        command_send_future.add_done_callback(did_send_command)

        # コマンドが '?' で終わるなら、レスポンスを期待する
        is_command_require_response = command.endswith("?")
        if is_command_require_response:
            read_response_future: Future = tmc_command_executor.submit(instrument.read_raw)
            read_response_future.add_done_callback(did_receive_response)

    # 終了
    print("Closing...")
    tmc_command_executor.shutdown()
    instrument.close()
    return 0


if __name__ == "__main__":
    result: int = main() or 0
    sys.exit(result)
